/*
             LUFA Library
     Copyright (C) Dean Camera, 2009.
              
  dean [at] fourwalledcubicle [dot] com
      www.fourwalledcubicle.com
*/

/*
  Copyright 2009  Dean Camera (dean [at] fourwalledcubicle [dot] com)

  Permission to use, copy, modify, and distribute this software
  and its documentation for any purpose and without fee is hereby
  granted, provided that the above copyright notice appear in all
  copies and that both that the copyright notice and this
  permission notice and warranty disclaimer appear in supporting
  documentation, and that the name of the author not be used in
  advertising or publicity pertaining to distribution of the
  software without specific, written prior permission.

  The author disclaim all warranties with regard to this
  software, including all implied warranties of merchantability
  and fitness.  In no event shall the author be liable for any
  special, indirect or consequential damages or any damages
  whatsoever resulting from loss of use, data or profits, whether
  in an action of contract, negligence or other tortious action,
  arising out of or in connection with the use or performance of
  this software.
*/

/************************************************************************/
/* THIS IS AN ADAPTED VERSION OF THE ORIGINAL LUFA CODE BY DEAN CAMERA  */
/* THE APPLIED CHANGES ARE NECESSARY TO MAKE THE CODE WORK FOR SERIAL   */
/* COMMUNICATION WITH THE EMU EMAX AND EMULATOR-II SAMPLERS VIA RS422   */
/*                                                                      */
/* ORIGINAL FILENAME WAS: USBToSerial.c                                 */
/* ORIGINAL LUFA RELEASE WAS: LUFA 090924 (24 SEPT 2009)                */
/* ORIGINAL LUFA FOLDER WAS: Demo\Device\Lowlevel\USBToSerial           */
/*                                                                      */
/* CHANGES APPLIED BY JULIAN HIGGINSON AND ESYNTHESIST, DECEMBER 2009   */
/* ADDITIONAL CHANGES BY ESYNTHESIST:                                   */
/* 2016-05-05: EMUSER V1.01.1: SUPPORT FOR POSIX-ONLY BAUDRATES         */
/* 2016-07-02: EMUSER V1.02.1: EXTENDED POSIX SUPPORT + OTHER LED MODE  */
/************************************************************************/


#include "USBtoSerialEmu.h"

/* Globals: */
/** Contains the current baud rate and other settings of the virtual serial port.
 *
 *  These values are set by the host via a class-specific request, and the physical USART should be reconfigured to match the
 *  new settings each time they are changed by the host.
 */
CDC_Line_Coding_t LineEncoding = { .BaudRateBPS = 0,
                                   .CharFormat  = OneStopBit,
                                   .ParityType  = Parity_None,
                                   .DataBits    = 8            };

/** Ring (circular) buffer to hold the RX data - data from the host to the attached device on the serial port. */
RingBuff_t Rx_Buffer;

/** Ring (circular) buffer to hold the TX data - data from the attached device on the serial port to the host. */
RingBuff_t Tx_Buffer;

/** Flag to indicate if the USART is currently transmitting data from the Rx_Buffer circular buffer. */
volatile bool Transmitting = false;

/*** EMUSER V1.02.1: SERIAL CONFIGURATION INDICATED BY LED - ADDED BY E-SYNTHESIST 2016-07-02 --> ***/
int  LEDMode = 3;   //serial config mode: 0=non-Emu mode, 1=Emu slow (midi) mode, 2=Emu fast (ext clock) mode, 3=zero bps mode
int  LEDFlashDone = 0;
long LEDSwitchCount[4], LEDCounter, LEDStatus, LEDCounterLimit;
/*** <-- ***/

/** Main program entry point. This routine configures the hardware required by the application, then
 *  starts the scheduler to run the application tasks.
 */
int main(void)
{
	SetupHardware();

	/* Ring buffer Initialization */
	Buffer_Initialize(&Rx_Buffer);
	Buffer_Initialize(&Tx_Buffer);

	LEDs_SetAllLEDs(LEDMASK_USB_NOTREADY);

	SetupLEDFlash();  //EMUSER V1.02.1: ADDED BY ESYNTHESIST 2016-07-02
	
	for (;;)
	{
		CDC_Task();
		USB_USBTask();
		if (LEDFlashDone == 0)   //EMUSER V1.02.1: ADDED BY ESYNTHESIST 2016-07-02
		  HandleLEDFlash();
	}
}

/** Configures the board hardware and chip peripherals for the demo's functionality. */
void SetupHardware(void)
{
	/* Disable watchdog if enabled by bootloader/fuses */
	MCUSR &= ~(1 << WDRF);
	wdt_disable();

	/* Disable clock division */
	clock_prescale_set(clock_div_1);

	/* Hardware Initialization */
	Serial_Init(9600, false);
	LEDs_Init();
	USB_Init();
}

/** Event handler for the USB_Connect event. This indicates that the device is enumerating via the status LEDs and
 *  starts the library USB task to begin the enumeration and USB management process.
 */
void EVENT_USB_Device_Connect(void)
{
	/* Indicate USB enumerating */

	LEDs_SetAllLEDs(LEDMASK_USB_ENUMERATING);
}

/** Event handler for the USB_Disconnect event. This indicates that the device is no longer connected to a host via
 *  the status LEDs and stops the USB management and CDC management tasks.
 */
void EVENT_USB_Device_Disconnect(void)
{	
	/* Reset Tx and Rx buffers, device disconnected */
	Buffer_Initialize(&Rx_Buffer);
	Buffer_Initialize(&Tx_Buffer);

	/* Indicate USB not ready */
	LEDs_SetAllLEDs(LEDMASK_USB_NOTREADY);
}

/** Event handler for the USB_ConfigurationChanged event. This is fired when the host set the current configuration
 *  of the USB device after enumeration - the device endpoints are configured and the CDC management task started.
 */
void EVENT_USB_Device_ConfigurationChanged(void)
{
	/* Indicate USB connected and ready */
	LEDs_SetAllLEDs(LEDMASK_USB_READY);

	/* Setup CDC Notification, Rx and Tx Endpoints */
	if (!(Endpoint_ConfigureEndpoint(CDC_NOTIFICATION_EPNUM, EP_TYPE_INTERRUPT,
		                             ENDPOINT_DIR_IN, CDC_NOTIFICATION_EPSIZE,
	                                 ENDPOINT_BANK_SINGLE)))
	{
		LEDs_SetAllLEDs(LEDMASK_USB_ERROR);
	}
	
	if (!(Endpoint_ConfigureEndpoint(CDC_TX_EPNUM, EP_TYPE_BULK,
		                             ENDPOINT_DIR_IN, CDC_TX_EPSIZE, //KRIS was: CDC_TXRX_EPSIZE
	                                 ENDPOINT_BANK_SINGLE)))
	{
		LEDs_SetAllLEDs(LEDMASK_USB_ERROR);
	}							   

	if (!(Endpoint_ConfigureEndpoint(CDC_RX_EPNUM, EP_TYPE_BULK,
		                             ENDPOINT_DIR_OUT, CDC_RX_EPSIZE,  //KRIS was: CDC_TXRX_EPSIZE
	                                 ENDPOINT_BANK_SINGLE)))
	{
		LEDs_SetAllLEDs(LEDMASK_USB_ERROR);
	}

	/* Reset line encoding baud rate so that the host knows to send new values */
	LineEncoding.BaudRateBPS = 0;
}

/** Event handler for the USB_UnhandledControlRequest event. This is used to catch standard and class specific
 *  control requests that are not handled internally by the USB library (including the CDC control commands,
 *  which are all issued via the control endpoint), so that they can be handled appropriately for the application.
 */
void EVENT_USB_Device_UnhandledControlRequest(void)
{
	/* Process CDC specific control requests */
	switch (USB_ControlRequest.bRequest)
	{
		case REQ_GetLineEncoding:
			if (USB_ControlRequest.bmRequestType == (REQDIR_DEVICETOHOST | REQTYPE_CLASS | REQREC_INTERFACE))
			{	
				/* Acknowledge the SETUP packet, ready for data transfer */
				Endpoint_ClearSETUP();

				/* Write the line coding data to the control endpoint */
				Endpoint_Write_Control_Stream_LE(&LineEncoding, sizeof(LineEncoding));
				
				/* Finalize the stream transfer to send the last packet or clear the host abort */
				Endpoint_ClearOUT();
			}
			
			break;
		case REQ_SetLineEncoding:
			if (USB_ControlRequest.bmRequestType == (REQDIR_HOSTTODEVICE | REQTYPE_CLASS | REQREC_INTERFACE))
			{
				/* Acknowledge the SETUP packet, ready for data transfer */
				Endpoint_ClearSETUP();

				/* Read the line coding data in from the host into the global struct */
				Endpoint_Read_Control_Stream_LE(&LineEncoding, sizeof(LineEncoding));

				/* Finalize the stream transfer to clear the last packet from the host */
				Endpoint_ClearIN();
				
				/* Reconfigure the USART with the new settings */
				ReconfigureUSART();
			}
	
			break;
		case REQ_SetControlLineState:
			if (USB_ControlRequest.bmRequestType == (REQDIR_HOSTTODEVICE | REQTYPE_CLASS | REQREC_INTERFACE))
			{				
				/* Acknowledge the SETUP packet, ready for data transfer */
				Endpoint_ClearSETUP();
				
				/* NOTE: Here you can read in the line state mask from the host, to get the current state of the output handshake
				         lines. The mask is read in from the wValue parameter in USB_ControlRequest, and can be masked against the
						 CONTROL_LINE_OUT_* masks to determine the RTS and DTR line states using the following code:
				*/

				Endpoint_ClearStatusStage();
			}
	
			break;
	}
}

/** Task to manage CDC data transmission and reception to and from the host, from and to the physical USART. */
void CDC_Task(void)
{
	/* Device must be connected and configured for the task to run */
	if (USB_DeviceState != DEVICE_STATE_Configured)
	  return;
	  
#if 0
	/* NOTE: Here you can use the notification endpoint to send back line state changes to the host, for the special RS-232
			 handshake signal lines (and some error states), via the CONTROL_LINE_IN_* masks and the following code:
	*/

	USB_Notification_Header_t Notification = (USB_Notification_Header_t)
		{
			.NotificationType = (REQDIR_DEVICETOHOST | REQTYPE_CLASS | REQREC_INTERFACE),
			.Notification     = NOTIF_SerialState,
			.wValue           = 0,
			.wIndex           = 0,
			.wLength          = sizeof(uint16_t),
		};
		
	uint16_t LineStateMask;
	
	// Set LineStateMask here to a mask of CONTROL_LINE_IN_* masks to set the input handshake line states to send to the host
	
	Endpoint_SelectEndpoint(CDC_NOTIFICATION_EPNUM);
	Endpoint_Write_Stream_LE(&Notification, sizeof(Notification));
	Endpoint_Write_Stream_LE(&LineStateMask, sizeof(LineStateMask));
	Endpoint_ClearIN();
#endif

	/* Select the Serial Rx Endpoint */
	Endpoint_SelectEndpoint(CDC_RX_EPNUM);
	
	/* Check to see if a packet has been received from the host */
	if (Endpoint_IsOUTReceived())
	{
		/* Read the bytes in from the endpoint into the buffer while space is available */
		while (Endpoint_BytesInEndpoint() && (Rx_Buffer.Elements != BUFF_STATICSIZE))
		{
			/* Store each character from the endpoint */
			Buffer_StoreElement(&Rx_Buffer, Endpoint_Read_Byte());
		}
		
		/* Check to see if all bytes in the current packet have been read */
		if (!(Endpoint_BytesInEndpoint()))
		{
			/* Clear the endpoint buffer */
			Endpoint_ClearOUT();
		}
	}
	
	/* Check if Rx buffer contains data - if so, send it */
	if (Rx_Buffer.Elements)
	  Serial_TxByte(Buffer_GetElement(&Rx_Buffer));

	/* Select the Serial Tx Endpoint */
	Endpoint_SelectEndpoint(CDC_TX_EPNUM);

	/* Check if the Tx buffer contains anything to be sent to the host */

	/* experimental code
	while ( (UCSR1A & (1<<RXC1)) && LineEncoding.BaudRateBPS )
	{
		Counter = 0;
		while ( (UCSR1A & (1<<RXC1)) && LineEncoding.BaudRateBPS && (Counter < CDC_TX_EPSIZE) )
		{
 			ReceivedByte = UDR1;

			if ((USB_DeviceState == DEVICE_STATE_Configured) && LineEncoding.BaudRateBPS)
			{
	 			 Buffer_StoreElement(&Tx_Buffer, ReceivedByte);
				 Counter++;
			}
		}
	*/

	if ((Tx_Buffer.Elements) && LineEncoding.BaudRateBPS)
	{
		/* Wait until Serial Tx Endpoint Ready for Read/Write */
		Endpoint_WaitUntilReady();
		
		/* Write the bytes from the buffer to the endpoint while space is available */
		while (Tx_Buffer.Elements && Endpoint_IsReadWriteAllowed())
		{
			/* Write each byte retreived from the buffer to the endpoint */
			Endpoint_Write_Byte(Buffer_GetElement(&Tx_Buffer));
		}
		
		/* Remember if the packet to send completely fills the endpoint */
		bool IsFull = (Endpoint_BytesInEndpoint() == CDC_TX_EPSIZE);             /*** ADAPTED BY ESYNTHESIST - original code was: == CDC_TXRX_EPSIZE ***/
		
		/* Send the data */
		Endpoint_ClearIN();

		/* If no more data to send and the last packet filled the endpoint, send an empty packet to release
		 * the buffer on the receiver (otherwise all data will be cached until a non-full packet is received) */
		if (IsFull && !(Tx_Buffer.Elements))
		{
			/* Wait until Serial Tx Endpoint Ready for Read/Write */
			Endpoint_WaitUntilReady();
				
			/* Send an empty packet to terminate the transfer */
			Endpoint_ClearIN();
		}
	}
	/* experimental code
	}
	*/
}

/** ISR to handle the USART receive complete interrupt, fired each time the USART has received a character. This stores the received
 *  character into the Tx_Buffer circular buffer for later transmission to the host.
 */
ISR(USART1_RX_vect, ISR_BLOCK)
{
	uint8_t ReceivedByte = UDR1;

	/* Only store received characters if the USB interface is connected */
	if ((USB_DeviceState == DEVICE_STATE_Configured) && LineEncoding.BaudRateBPS)
	  Buffer_StoreElement(&Tx_Buffer, ReceivedByte);
	//put above 2 lines in comment for experimental code
}

void ClockDivider(uint32_t Fct)  /* EMUSER V1.02.1: clockdivider */
{
  if (Fct != 0)
    LineEncoding.BaudRateBPS*=Fct;
  return;
}

/** Reconfigures the USART to match the current serial port settings issued by the host as closely as possible. */
void ReconfigureUSART(void)
{
	uint8_t ConfigMask = 0;

	/*** EMUSER V1.01.1 and V1.02.1: POSIX COMPLIANCE AND LED HANDLING - CALL SUBROUTINE - ADDED BY E-SYNTHESIST 2016-05-05 & 2016-07-02 ***/

    ConfigurePosixLineEncodingAndLED();
    SetupLEDFlash();

    /*** <-- ***/	

	/* Determine parity - non odd/even parity mode defaults to no parity */
	if (LineEncoding.ParityType == Parity_Odd)
	  ConfigMask = ((1 << UPM11) | (1 << UPM10));
	else if (LineEncoding.ParityType == Parity_Even)
	  ConfigMask = (1 << UPM11);

	/* Determine stop bits - 1.5 stop bits is set as 1 stop bit due to hardware limitations */
	if (LineEncoding.CharFormat == TwoStopBits)
	  ConfigMask |= (1 << USBS1);

	/* Determine data size - 5, 6, 7, or 8 bits are supported */
	if (LineEncoding.DataBits == 6)
	  ConfigMask |= (1 << UCSZ10);
	else if (LineEncoding.DataBits == 7)
	  ConfigMask |= (1 << UCSZ11);
	else if (LineEncoding.DataBits == 8)
	  ConfigMask |= ((1 << UCSZ11) | (1 << UCSZ10));

    /*** ADDED BY JULIAN --> ***/

	if (LineEncoding.BaudRateBPS == 500000)
	{

		// UMSEL11,UMSEL10 -> 0,1 for synchronous mode. UCPOL1 -> 1 for recieved data sampled on RISING edge
		ConfigMask |= ((0<<UMSEL11) | (1<<UMSEL10) | (1 << UCPOL1));

		// set clock input/output
		DDRD &= ~(1<<5);	

		/* Enable double speed, gives better error percentages at 8MHz */
		UCSR1A = (0 << U2X1);
	
		/* Set the USART mode to the mask generated by the Line Coding options */
		UCSR1C = ConfigMask;
	
		/* Enable transmit and receive modules and interrupts */
		UCSR1B = ((1 << RXCIE1) | (1 << TXEN1) | (1 << RXEN1));
	}
	else
	{
	 	// Set clock input/output
		DDRD |= (1<<5);

		// Enable double speed, gives better error percentages at 8MHz
		UCSR1A = (1 << U2X1);

		// Set the USART mode to the mask generated by the Line Coding options
		UCSR1C = ConfigMask;
	
		// Set the USART baud rate register to the desired baud rate value
		UBRR1L  = (unsigned char)(SERIAL_2X_UBBRVAL(LineEncoding.BaudRateBPS));
		UBRR1H  = (unsigned char)((SERIAL_2X_UBBRVAL(LineEncoding.BaudRateBPS)) >> 8);

		// Enable transmit and receive modules and interrupts
		UCSR1B = ((1 << RXCIE1) | (1 << TXEN1) | (1 << RXEN1));
	}

    /*** <-- ***/

    /*** ORIGINAL LUFA PROJECT CODE IN COMMENTS FOLLOWS HERE -->
	Enable double speed, gives better error percentages at 8MHz
	UCSR1A = (1 << U2X1);
	
	Enable transmit and receive modules and interrupts
	UCSR1B = ((1 << RXCIE1) | (1 << TXEN1) | (1 << RXEN1));

	Set the USART mode to the mask generated by the Line Coding options
	UCSR1C = ConfigMask;
	
	Set the USART baud rate register to the desired baud rate value
	UBRR1  = SERIAL_2X_UBBRVAL((uint16_t)LineEncoding.BaudRateBPS);
    <-- ***/


}

void HandleLEDFlash (void)
{
  //EMUSER V1.02.1: flash the LED twice if a non-EMU or non-zero baud rate configuration is requested -- ADDED BY E-ESYNTHESIST 2016-07-02

  LEDCounter++;

  if (LEDCounter > LEDCounterLimit)
    LEDCounter = 1;

  //Note: the test below uses 4 discrete test values instead of algorithmic code in order to optimize speed
  if ( (LEDCounter == LEDSwitchCount[0]) || (LEDCounter == LEDSwitchCount[1]) || (LEDCounter == LEDSwitchCount[2]) || (LEDCounter == LEDSwitchCount[3]) )
  {
    if (LEDStatus == 1)
	{
	  LEDs_TurnOffLEDs(LEDMASK_SER_CONFIG);
	  LEDStatus = 0;
    }
	else
	{
	  LEDs_TurnOnLEDs(LEDMASK_SER_CONFIG);
	  LEDStatus = 1;
    }
  }
  else
    if (LEDCounter > LEDSwitchCount[3])
    {
      LEDFlashDone = 1;
	  LEDs_TurnOnLEDs(LEDMASK_SER_CONFIG);
    }

  return;
}

void SetupLEDFlash (void)
{
  /* EMUSER V1.02.1: set LED handling to indicate selected serial configuration -- ADDED BY E-ESYNTHESIST 2016-07-02
     These are the modes:
	 LEDMode = 0: flash the LED twice then keep LED ON when any configuration other than the 3 modes below is selected
	 LEDMode = 1: turn OFF LED when MIDI speed is requested
	 LEDMode = 2: turn ON  LED when external clock is requested
	 LEDMode = 3: turn ON  LED when a reset of the BPS is done (=device ready mode)
  */ 
  int  i;

  LEDCounter = 1;
  LEDStatus = 1;
  LEDCounterLimit = (long)F_CPU/16;
  LEDFlashDone = 0;
  for(i=0;i<4;i++)
    LEDSwitchCount[i] = LEDCounterLimit+1;

  switch(LEDMode)
  {
    case 1:  { LEDs_TurnOffLEDs(LEDMASK_SER_CONFIG); LEDFlashDone = 1; break; }   //slow baud rate
	case 2:
	case 3:  { LEDs_TurnOnLEDs(LEDMASK_SER_CONFIG); LEDFlashDone = 1; break; }    //high speed/ext clock and zero baud rate
	default: { LEDs_TurnOnLEDs(LEDMASK_SER_CONFIG);
	           for(i=0;i<4;i++)
			     LEDSwitchCount[i] = (long)(((long)F_CPU/3200)*(i+1)); //other baud rate. Flash 4 times. Fast flashing = every F_CPU/6400, slow flashing = every F_CPU/1280
               break;
			 }
  }

  return;
}

void ConfigurePosixLineEncodingAndLED (void)
{
	/*** EMUSER V1.01.1 and V1.02.1: FOR COMPATIBILITY WITH POSIX STANDARD - ADDED BY E-SYNTHESIST 2016-05-05 & 2016-07-02
         EMUSER V1.02.1: ADD LED SET UP FOR INDICATING SELECTED SPEED - ADDED BY E-SYNTHESSIST 2016-07-02
 	 --> ***/
  uint32_t Stpb;
  uint8_t  Mode;

  Stpb = SetModeAndPosixStpb(&Mode);

  if ( (LineEncoding.BaudRateBPS == 50) && (Mode == ModeFast) )
  {
	  LineEncoding.BaudRateBPS = 500000;
	  LineEncoding.DataBits = 8;
	  LineEncoding.CharFormat = OneStopBit;
	  LineEncoding.ParityType = Parity_Odd;
  }
  else
    if ( (LineEncoding.BaudRateBPS == 38400) && (Mode == ModeSlow) )
	{
	  LineEncoding.BaudRateBPS = 31250;
	  LineEncoding.DataBits = 8;
	  LineEncoding.CharFormat = OneStopBit;
	  LineEncoding.ParityType = Parity_None;
    }

  if ( (LineEncoding.BaudRateBPS == 500000) && (Mode == ModeFast) )
    LEDMode = 2;
  else
	  if ( (LineEncoding.BaudRateBPS == 31250) && (Mode == ModeSlow) )
		LEDMode = 1;
	  else
	    if (LineEncoding.BaudRateBPS == 0)
		  LEDMode = 3;
        else
		  LEDMode = 0;

  ClockDivider(Stpb);

  return;
}

uint32_t SetModeAndPosixStpb (uint8_t * Mode)
{
  /*** EMUSER V1.02.1 ***/

  uint32_t Stpb;
  uint8_t  xMode;

  Stpb = PosixStpb0n;
  xMode = ModeNormal;

  if (LineEncoding.DataBits == 8)
  {
    switch(LineEncoding.CharFormat)
    {
     case OneStopBit:
	 {
	  switch(LineEncoding.ParityType)
	  {
	    case Parity_Odd:   { Stpb = PosixStpb1o; xMode = ModeFast; break; }
        case Parity_Even:  { Stpb = PosixStpb1e; xMode = ModeNormal; break; }
		case Parity_None:  { Stpb = PosixStpb1n; xMode = ModeSlow; break; }
	  }
	  break;
     }
	 case TwoStopBits:
	 {
	  switch(LineEncoding.ParityType)
	  {
	    case Parity_Odd:   { Stpb = PosixStpb2o; xMode = ModeNormal; break; }
        case Parity_Even:  { Stpb = PosixStpb2e; xMode = ModeSlow; break; }
		case Parity_None:  { Stpb = PosixStpb2n; xMode = ModeNormal; break; }
	  }
	  break;
     }
    }
  }
 
  if (Mode != NULL)
    *Mode = xMode;

  return(Stpb);
}

